from gitlint.rules import CommitRule, RuleViolation


class SignedOffBy(CommitRule):
    """This rule will enforce that each commit body contains a "Signed-off-by" line.
    We keep things simple here and just check whether the commit body contains a line that starts with "Signed-off-by".
    """

    name = "contrib-body-requires-signed-off-by"
    id = "CC1"

    def validate(self, commit):
        for line in commit.message.body:
            if line.lower().startswith("signed-off-by"):
                return []

        return [RuleViolation(self.id, "Body does not contain a 'Signed-off-by' line", line_nr=1)]



from gitlint.rules import CommitRule, RuleViolation


class DisallowCleanupCommits(CommitRule):
    """This rule checks the commits for "fixup!"/"squash!"/"amend!" commits
    and rejects them.
    """

    name = "contrib-disallow-cleanup-commits"
    id = "CC2"

    def validate(self, commit):
        if commit.is_fixup_commit:
            return [RuleViolation(self.id, "Fixup commits are not allowed", line_nr=1)]

        if commit.is_squash_commit:
            return [RuleViolation(self.id, "Squash commits are not allowed", line_nr=1)]

        if commit.is_fixup_amend_commit:
            return [RuleViolation(self.id, "Amend commits are not allowed", line_nr=1)]

        return []



import copy
import logging
import os
import platform
import stat
import sys

import click

import gitlint
from gitlint import hooks
from gitlint.config import LintConfigBuilder, LintConfigError, LintConfigGenerator
from gitlint.deprecation import DEPRECATED_LOG_FORMAT
from gitlint.deprecation import LOG as DEPRECATED_LOG
from gitlint.exception import GitlintError
from gitlint.git import GitContext, GitContextError, git_version
from gitlint.lint import GitLinter
from gitlint.shell import shell
from gitlint.utils import LOG_FORMAT

# Error codes
GITLINT_SUCCESS = 0
MAX_VIOLATION_ERROR_CODE = 252
USAGE_ERROR_CODE = 253
GIT_CONTEXT_ERROR_CODE = 254
CONFIG_ERROR_CODE = 255

DEFAULT_CONFIG_FILE = ".gitlint"
# -n: disable swap files. This fixes a vim error on windows (E303: Unable to open swap file for <path>)
DEFAULT_COMMIT_MSG_EDITOR = "vim -n"

# Since we use the return code to denote the amount of errors, we need to change the default click usage error code
click.UsageError.exit_code = USAGE_ERROR_CODE

# We don't use logging.getLogger(__main__) here because that will cause DEBUG output to be lost
# when invoking gitlint as a python module (python -m gitlint.cli)
LOG = logging.getLogger("gitlint.cli")


class GitLintUsageError(GitlintError):
    """Exception indicating there is an issue with how gitlint is used."""


def setup_logging():
    """Setup gitlint logging"""

    # Root log, mostly used for debug
    root_log = logging.getLogger("gitlint")
    root_log.propagate = False  # Don't propagate to child loggers, the gitlint root logger handles everything
    root_log.setLevel(logging.WARN)
    handler = logging.StreamHandler()
    formatter = logging.Formatter(LOG_FORMAT)
    handler.setFormatter(formatter)
    root_log.addHandler(handler)

    # Deprecated log, to log deprecation warnings
    DEPRECATED_LOG.propagate = False  # Don't propagate to child logger
    DEPRECATED_LOG.setLevel(logging.WARNING)
    deprecated_log_handler = logging.StreamHandler()
    deprecated_log_handler.setFormatter(logging.Formatter(DEPRECATED_LOG_FORMAT))
    DEPRECATED_LOG.addHandler(deprecated_log_handler)


def log_system_info():
    LOG.debug("Platform: %s", platform.platform())
    LOG.debug("Python version: %s", sys.version)
    LOG.debug("Git version: %s", git_version())
    LOG.debug("Gitlint version: %s", gitlint.__version__)
    LOG.debug("GITLINT_USE_SH_LIB: %s", os.environ.get("GITLINT_USE_SH_LIB", "[NOT SET]"))
    LOG.debug("TERMINAL_ENCODING: %s", gitlint.utils.TERMINAL_ENCODING)
    LOG.debug("FILE_ENCODING: %s", gitlint.utils.FILE_ENCODING)


def build_config(
    target,
    config_path,
    c,
    extra_path,
    ignore,
    contrib,
    ignore_stdin,
    staged,
    fail_without_commits,
    verbose,
    silent,
    debug,
):
    """Creates a LintConfig object based on a set of commandline parameters."""
    config_builder = LintConfigBuilder()
    # Config precedence:
    # First, load default config or config from configfile
    if config_path:
        config_builder.set_from_config_file(config_path)
    elif os.path.exists(DEFAULT_CONFIG_FILE):
        config_builder.set_from_config_file(DEFAULT_CONFIG_FILE)

    # Then process any commandline configuration flags
    config_builder.set_config_from_string_list(c)

    # Finally, overwrite with any convenience commandline flags
    if ignore:
        config_builder.set_option("general", "ignore", ignore)

    if contrib:
        config_builder.set_option("general", "contrib", contrib)

    if ignore_stdin:
        config_builder.set_option("general", "ignore-stdin", ignore_stdin)

    if silent:
        config_builder.set_option("general", "verbosity", 0)
    elif verbose > 0:
        config_builder.set_option("general", "verbosity", verbose)

    if extra_path:
        config_builder.set_option("general", "extra-path", extra_path)

    if target:
        config_builder.set_option("general", "target", target)

    if debug:
        config_builder.set_option("general", "debug", debug)

    if staged:
        config_builder.set_option("general", "staged", staged)

    if fail_without_commits:
        config_builder.set_option("general", "fail-without-commits", fail_without_commits)

    config = config_builder.build()

    return config, config_builder


def get_stdin_data():
    """Helper function that returns data sent to stdin or False if nothing is sent"""
    # STDIN can only be 3 different types of things ("modes")
    #  1. An interactive terminal device (i.e. a TTY -> sys.stdin.isatty() or stat.S_ISCHR)
    #  2. A (named) pipe (stat.S_ISFIFO)
    #  3. A regular file (stat.S_ISREG)
    # Technically, STDIN can also be other device type like a named unix socket (stat.S_ISSOCK), but we don't
    # support that in gitlint (at least not today).
    #
    # Now, the behavior that we want is the following:
    # If someone sends something directly to gitlint via a pipe or a regular file, read it. If not, read from the
    # local repository.
    # Note that we don't care about whether STDIN is a TTY or not, we only care whether data is via a pipe or regular
    # file.
    # However, in case STDIN is not a TTY, it HAS to be one of the 2 other things (pipe or regular file), even if
    # no-one is actually sending anything to gitlint over them. In this case, we still want to read from the local
    # repository.
    # To support this use-case (which is common in CI runners such as Jenkins and Gitlab), we need to actually attempt
    # to read from STDIN in case it's a pipe or regular file. In case that fails, then we'll fall back to reading
    # from the local repo.

    mode = os.fstat(sys.stdin.fileno()).st_mode
    stdin_is_pipe_or_file = stat.S_ISFIFO(mode) or stat.S_ISREG(mode)
    if stdin_is_pipe_or_file:
        input_data = sys.stdin.read()
        # Only return the input data if there's actually something passed
        # i.e. don't consider empty piped data
        if input_data:
            return str(input_data)
    return False


def build_git_context(lint_config, msg_filename, commit_hash, refspec):
    """Builds a git context based on passed parameters and order of precedence"""

    # Determine which GitContext method to use if a custom message is passed
    from_commit_msg = GitContext.from_commit_msg
    if lint_config.staged:
        LOG.debug("Fetching additional meta-data from staged commit")

        def from_commit_msg(message):
            return GitContext.from_staged_commit(message, lint_config.target)

    # Order of precedence:
    # 1. Any data specified via --msg-filename
    if msg_filename:
        LOG.debug("Using --msg-filename.")
        return from_commit_msg(str(msg_filename.read()))

    # 2. Any data sent to stdin (unless stdin is being ignored)
    if not lint_config.ignore_stdin:
        stdin_input = get_stdin_data()
        if stdin_input:
            LOG.debug("Stdin data: '%s'", stdin_input)
            LOG.debug("Stdin detected and not ignored. Using as input.")
            return from_commit_msg(stdin_input)

    if lint_config.staged:
        raise GitLintUsageError(
            "The 'staged' option (--staged) can only be used when using '--msg-filename' or "
            "when piping data to gitlint via stdin."
        )

    # 3. Fallback to reading from local repository
    LOG.debug("No --msg-filename flag, no or empty data passed to stdin. Using the local repo.")

    if commit_hash and refspec:
        raise GitLintUsageError("--commit and --commits are mutually exclusive, use one or the other.")

    # 3.1 Linting a range of commits
    if refspec:
        # 3.1.1 Not real refspec, but comma-separated list of commit hashes
        if "," in refspec:
            commit_hashes = [hash.strip() for hash in refspec.split(",") if hash]
            return GitContext.from_local_repository(lint_config.target, commit_hashes=commit_hashes)
        # 3.1.2 Real refspec
        return GitContext.from_local_repository(lint_config.target, refspec=refspec)

    # 3.2 Linting a specific commit
    if commit_hash:
        return GitContext.from_local_repository(lint_config.target, commit_hashes=[commit_hash])

    # 3.3 Fallback to linting the current HEAD
    return GitContext.from_local_repository(lint_config.target)


def handle_gitlint_error(ctx, exc):
    """Helper function to handle exceptions"""
    if isinstance(exc, GitContextError):
        click.echo(exc)
        ctx.exit(GIT_CONTEXT_ERROR_CODE)
    elif isinstance(exc, GitLintUsageError):
        click.echo(f"Error: {exc}")
        ctx.exit(USAGE_ERROR_CODE)
    elif isinstance(exc, LintConfigError):
        click.echo(f"Config Error: {exc}")
        ctx.exit(CONFIG_ERROR_CODE)


class ContextObj:
    """Simple class to hold data that is passed between Click commands via the Click context."""

    def __init__(self, config, config_builder, commit_hash, refspec, msg_filename, gitcontext=None):
        self.config = config
        self.config_builder = config_builder
        self.commit_hash = commit_hash
        self.refspec = refspec
        self.msg_filename = msg_filename
        self.gitcontext = gitcontext


# fmt: off
@click.group(invoke_without_command=True, context_settings={"max_content_width": 120},
             epilog="When no COMMAND is specified, gitlint defaults to 'gitlint lint'.")
@click.option("--target", envvar="GITLINT_TARGET",
              type=click.Path(exists=True, resolve_path=True, file_okay=False, readable=True),
              help="Path of the target git repository. [default: current working directory]")
@click.option("-C", "--config", envvar="GITLINT_CONFIG",
              type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
              help=f"Config file location [default: {DEFAULT_CONFIG_FILE}]")
@click.option("-c", multiple=True,
              help="Config flags in format <rule>.<option>=<value> (e.g.: -c T1.line-length=80). " +
                   "Flag can be used multiple times to set multiple config values.")
@click.option("--commit", envvar="GITLINT_COMMIT", default=None, help="Hash (SHA) of specific commit to lint.")
@click.option("--commits", envvar="GITLINT_COMMITS", default=None,
              help="The range of commits (refspec or comma-separated hashes) to lint. [default: HEAD]")
@click.option("-e", "--extra-path", envvar="GITLINT_EXTRA_PATH",
              help="Path to a directory or python module with extra user-defined rules",
              type=click.Path(exists=True, resolve_path=True, readable=True))
@click.option("--ignore", envvar="GITLINT_IGNORE", default="", help="Ignore rules (comma-separated by id or name).")
@click.option("--contrib", envvar="GITLINT_CONTRIB", default="",
              help="Contrib rules to enable (comma-separated by id or name).")
@click.option("--msg-filename", type=click.File(encoding=gitlint.utils.FILE_ENCODING),
              help="Path to a file containing a commit-msg.")
@click.option("--ignore-stdin", envvar="GITLINT_IGNORE_STDIN", is_flag=True,
              help="Ignore any stdin data. Useful for running in CI server.")
@click.option("--staged", envvar="GITLINT_STAGED", is_flag=True,
              help="Attempt smart guesses about meta info (like author name, email, branch, changed files, etc) " +
                   "for staged commits.")
@click.option("--fail-without-commits", envvar="GITLINT_FAIL_WITHOUT_COMMITS", is_flag=True,
              help="Hard fail when the target commit range is empty.")
@click.option("-v", "--verbose", envvar="GITLINT_VERBOSITY", count=True, default=0,
              help="Verbosity, use multiple times for more verbose output (e.g.: -v, -vv, -vvv). [default: -vvv]", )
@click.option("-s", "--silent", envvar="GITLINT_SILENT", is_flag=True,
              help="Silent mode (no output). Takes precedence over -v, -vv, -vvv.")
@click.option("-d", "--debug", envvar="GITLINT_DEBUG", help="Enable debugging output.", is_flag=True)
@click.version_option(version=gitlint.__version__)
@click.pass_context
def cli(
        ctx, target, config, c, commit, commits, extra_path, ignore, contrib,
        msg_filename, ignore_stdin, staged, fail_without_commits, verbose,
        silent, debug,
):
    """ Git lint tool, checks your git commit messages for styling issues

        Documentation: http://jorisroovers.github.io/gitlint
    """
    try:
        if debug:
            logging.getLogger("gitlint").setLevel(logging.DEBUG)
            DEPRECATED_LOG.setLevel(logging.DEBUG)
        LOG.debug("To report issues, please visit https://github.com/jorisroovers/gitlint/issues")

        log_system_info()

        # Get the lint config from the commandline parameters and
        # store it in the context (click allows storing an arbitrary object in ctx.obj).
        config, config_builder = build_config(target, config, c, extra_path, ignore, contrib, ignore_stdin,
                                              staged, fail_without_commits, verbose, silent, debug)
        LOG.debug("Configuration\n%s", config)

        ctx.obj = ContextObj(config, config_builder, commit, commits, msg_filename)

        # If no subcommand is specified, then just lint
        if ctx.invoked_subcommand is None:
            ctx.invoke(lint)

    except GitlintError as e:
        handle_gitlint_error(ctx, e)
# fmt: on


@cli.command("lint")
@click.pass_context
def lint(ctx):
    """Lints a git repository [default command]"""
    lint_config = ctx.obj.config
    refspec = ctx.obj.refspec
    commit_hash = ctx.obj.commit_hash
    msg_filename = ctx.obj.msg_filename

    gitcontext = build_git_context(lint_config, msg_filename, commit_hash, refspec)
    # Set gitcontext in the click context, so we can use it in command that are ran after this
    # in particular, this is used by run-hook
    ctx.obj.gitcontext = gitcontext

    number_of_commits = len(gitcontext.commits)
    # Exit if we don't have commits in the specified range. Use a 0 exit code, since a popular use-case is one
    # where users are using --commits in a check job to check the commit messages inside a CI job. By returning 0, we
    # ensure that these jobs don't fail if for whatever reason the specified commit range is empty.
    # This behavior can be overridden by using the --fail-without-commits flag.
    if number_of_commits == 0:
        LOG.debug('No commits in range "%s"', refspec)
        if lint_config.fail_without_commits:
            raise GitLintUsageError(f'No commits in range "{refspec}"')
        ctx.exit(GITLINT_SUCCESS)

    LOG.debug("Linting %d commit(s)", number_of_commits)
    general_config_builder = ctx.obj.config_builder
    last_commit = gitcontext.commits[-1]

    # Let's get linting!
    first_violation = True
    exit_code = GITLINT_SUCCESS
    for commit in gitcontext.commits:
        # Build a config_builder taking into account the commit specific config (if any)
        config_builder = general_config_builder.clone()
        config_builder.set_config_from_commit(commit)

        # Create a deepcopy from the original config, so we have a unique config object per commit
        # This is important for configuration rules to be able to modifying the config on a per commit basis
        commit_config = config_builder.build(copy.deepcopy(lint_config))

        # Actually do the linting
        linter = GitLinter(commit_config)
        violations = linter.lint(commit)
        # exit code equals the total number of violations in all commits
        exit_code += len(violations)
        if violations:
            # Display the commit hash & new lines intelligently
            if number_of_commits > 1 and commit.sha:
                commit_separator = "\n" if not first_violation or commit is last_commit else ""
                linter.display.e(f"{commit_separator}Commit {commit.sha[:10]}:")
            linter.print_violations(violations)
            first_violation = False

    # cap actual max exit code because bash doesn't like exit codes larger than 255:
    # http://tldp.org/LDP/abs/html/exitcodes.html
    exit_code = min(MAX_VIOLATION_ERROR_CODE, exit_code)
    LOG.debug("Exit Code = %s", exit_code)
    ctx.exit(exit_code)


@cli.command("install-hook")
@click.pass_context
def install_hook(ctx):
    """Install gitlint as a git commit-msg hook."""
    try:
        hooks.GitHookInstaller.install_commit_msg_hook(ctx.obj.config)
        hook_path = hooks.GitHookInstaller.commit_msg_hook_path(ctx.obj.config)
        click.echo(f"Successfully installed gitlint commit-msg hook in {hook_path}")
        ctx.exit(GITLINT_SUCCESS)
    except hooks.GitHookInstallerError as e:
        click.echo(e, err=True)
        ctx.exit(GIT_CONTEXT_ERROR_CODE)


@cli.command("uninstall-hook")
@click.pass_context
def uninstall_hook(ctx):
    """Uninstall gitlint commit-msg hook."""
    try:
        hooks.GitHookInstaller.uninstall_commit_msg_hook(ctx.obj.config)
        hook_path = hooks.GitHookInstaller.commit_msg_hook_path(ctx.obj.config)
        click.echo(f"Successfully uninstalled gitlint commit-msg hook from {hook_path}")
        ctx.exit(GITLINT_SUCCESS)
    except hooks.GitHookInstallerError as e:
        click.echo(e, err=True)
        ctx.exit(GIT_CONTEXT_ERROR_CODE)


@cli.command("run-hook")
@click.pass_context
def run_hook(ctx):
    """Runs the gitlint commit-msg hook."""

    exit_code = 1
    while exit_code > 0:
        try:
            click.echo("gitlint: checking commit message...")
            ctx.invoke(lint)
        except GitlintError as e:
            handle_gitlint_error(ctx, e)
        except click.exceptions.Exit as e:
            # Flush stderr andstdout, this resolves an issue with output ordering in Cygwin
            sys.stderr.flush()
            sys.stdout.flush()

            exit_code = e.exit_code
            if exit_code == GITLINT_SUCCESS:
                click.echo("gitlint: " + click.style("OK", fg="green") + " (no violations in commit message)")
                continue

            click.echo("-----------------------------------------------")
            click.echo("gitlint: " + click.style("Your commit message contains violations.", fg="red"))

            value = None
            while value not in ["y", "n", "e"]:
                click.echo(
                    "Continue with commit anyways (this keeps the current commit message)? [y(es)/n(no)/e(dit)] ",
                    nl=False,
                )

                # Ideally, we'd want to use click.getchar() or click.prompt() to get user's input here instead of
                # input(). However, those functions currently don't support getting answers from stdin.
                # This wouldn't be a huge issue since this is unlikely to occur in the real world,
                # were it not that we use a stdin to pipe answers into gitlint in our integration tests.
                # If that ever changes, we can revisit this.
                # Related click pointers:
                # - https://github.com/pallets/click/issues/1370
                # - https://github.com/pallets/click/pull/1372
                # - From https://click.palletsprojects.com/en/7.x/utils/#getting-characters-from-terminal
                #   Note that this function will always read from the terminal, even if stdin is instead a pipe.
                value = input()

            if value == "y":
                LOG.debug("run-hook: commit message accepted")
                exit_code = GITLINT_SUCCESS
            elif value == "e":
                LOG.debug("run-hook: editing commit message")
                msg_filename = ctx.obj.msg_filename
                if msg_filename:
                    msg_filename.seek(0)
                    editor = os.environ.get("EDITOR", DEFAULT_COMMIT_MSG_EDITOR)
                    msg_filename_path = os.path.realpath(msg_filename.name)
                    LOG.debug("run-hook: %s %s", editor, msg_filename_path)
                    shell(editor + " " + msg_filename_path)
                else:
                    click.echo("Editing only possible when --msg-filename is specified.")
                    ctx.exit(exit_code)
            elif value == "n":
                LOG.debug("run-hook: commit message declined")
                click.echo("Commit aborted.")
                click.echo("Your commit message: ")
                click.echo("-----------------------------------------------")
                click.echo(ctx.obj.gitcontext.commits[0].message.full)
                click.echo("-----------------------------------------------")
                ctx.exit(exit_code)

    ctx.exit(exit_code)


@cli.command("generate-config")
@click.pass_context
def generate_config(ctx):
    """Generates a sample gitlint config file."""
    path = click.prompt("Please specify a location for the sample gitlint config file", default=DEFAULT_CONFIG_FILE)
    path = os.path.realpath(path)
    dir_name = os.path.dirname(path)
    if not os.path.exists(dir_name):
        click.echo(f"Error: Directory '{dir_name}' does not exist.", err=True)
        ctx.exit(USAGE_ERROR_CODE)
    elif os.path.exists(path):
        click.echo(f'Error: File "{path}" already exists.', err=True)
        ctx.exit(USAGE_ERROR_CODE)

    LintConfigGenerator.generate_config(path)
    click.echo(f"Successfully generated {path}")
    ctx.exit(GITLINT_SUCCESS)


# Let's Party!
setup_logging()
if __name__ == "__main__":
    cli()  # pragma: no cover



import logging
import os
from pathlib import Path

import arrow

from gitlint import shell as sh
from gitlint.cache import PropertyCache, cache
from gitlint.exception import GitlintError

# import exceptions separately, this makes it a little easier to mock them out in the unit tests
from gitlint.shell import CommandNotFound, ErrorReturnCode

# For now, the git date format we use is fixed, but technically this format is determined by `git config log.date`
# We should fix this at some point :-)
GIT_TIMEFORMAT = "YYYY-MM-DD HH:mm:ss Z"

LOG = logging.getLogger(__name__)


class GitContextError(GitlintError):
    """Exception indicating there is an issue with the git context"""


class GitNotInstalledError(GitContextError):
    def __init__(self):
        super().__init__(
            "'git' command not found. You need to install git to use gitlint on a local repository. "
            "See https://git-scm.com/book/en/v2/Getting-Started-Installing-Git on how to install git."
        )


class GitExitCodeError(GitContextError):
    def __init__(self, command, stderr):
        self.command = command
        self.stderr = stderr
        super().__init__(f"An error occurred while executing '{command}': {stderr}")


def _git(*command_parts, **kwargs):
    """Convenience function for running git commands. Automatically deals with exceptions and unicode."""
    git_kwargs = {"_tty_out": False}
    git_kwargs.update(kwargs)
    try:
        LOG.debug(command_parts)
        result = sh.git(*command_parts, **git_kwargs)
        # If we reach this point and the result has an exit_code that is larger than 0, this means that we didn't
        # get an exception (which is the default sh behavior for non-zero exit codes) and so the user is expecting
        # a non-zero exit code -> just return the entire result
        if hasattr(result, "exit_code") and result.exit_code > 0:
            return result
        return str(result)
    except CommandNotFound as e:
        raise GitNotInstalledError from e
    except ErrorReturnCode as e:  # Something went wrong while executing the git command
        error_msg = e.stderr.strip()
        error_msg_lower = error_msg.lower()
        if "_cwd" in git_kwargs and b"not a git repository" in error_msg_lower:
            raise GitContextError(f"{git_kwargs['_cwd']} is not a git repository.") from e

        if (
            b"does not have any commits yet" in error_msg_lower
            or b"ambiguous argument 'head': unknown revision" in error_msg_lower
        ):
            msg = "Current branch has no commits. Gitlint requires at least one commit to function."
            raise GitContextError(msg) from e

        raise GitExitCodeError(e.full_cmd, error_msg) from e


def git_version():
    """Determine the git version installed on this host by calling git --version"""
    return _git("--version").replace("\n", "")


def git_commentchar(repository_path=None):
    """Shortcut for retrieving comment char from git config"""
    commentchar = _git("config", "--get", "core.commentchar", _cwd=repository_path, _ok_code=[0, 1])
    # git will return an exit code of 1 if it can't find a config value, in this case we fall-back to # as commentchar
    if hasattr(commentchar, "exit_code") and commentchar.exit_code == 1:
        commentchar = "#"
    return commentchar.replace("\n", "")


def git_hooks_dir(repository_path):
    """Determine hooks directory for a given target dir"""
    hooks_dir = _git("rev-parse", "--git-path", "hooks", _cwd=repository_path)
    hooks_dir = hooks_dir.replace("\n", "")
    return os.path.realpath(os.path.join(repository_path, hooks_dir))


def _parse_git_changed_file_stats(changed_files_stats_raw):
    """Parse the output of git diff --numstat and return a dict of:
    dict[filename: GitChangedFileStats(filename, additions, deletions)]"""
    changed_files_stats_lines = changed_files_stats_raw.split("\n")
    changed_files_stats = {}
    for line in changed_files_stats_lines[:-1]:  # drop last empty line
        line_stats = line.split()

        # If the file is binary, numstat will show "-"
        # See https://git-scm.com/docs/git-diff#Documentation/git-diff.txt---numstat
        additions = int(line_stats[0]) if line_stats[0] != "-" else None
        deletions = int(line_stats[1]) if line_stats[1] != "-" else None

        changed_file_stat = GitChangedFileStats(line_stats[2], additions, deletions)
        changed_files_stats[line_stats[2]] = changed_file_stat

    return changed_files_stats


class GitCommitMessage:
    """Class representing a git commit message. A commit message consists of the following:
    - context: The `GitContext` this commit message is part of
    - original: The actual commit message as returned by `git log`
    - full: original, but stripped of any comments
    - title: the first line of full
    - body: all lines following the title
    """

    def __init__(self, context, original=None, full=None, title=None, body=None):
        self.context = context
        self.original = original
        self.full = full
        self.title = title
        self.body = body

    @staticmethod
    def from_full_message(context, commit_msg_str):
        """Parses a full git commit message by parsing a given string into the different parts of a commit message"""
        all_lines = commit_msg_str.splitlines()
        cutline = f"{context.commentchar} ------------------------ >8 ------------------------"
        try:
            cutline_index = all_lines.index(cutline)
        except ValueError:
            cutline_index = None
        lines = [line for line in all_lines[:cutline_index] if not line.startswith(context.commentchar)]
        full = "\n".join(lines)
        title = lines[0] if lines else ""
        body = lines[1:] if len(lines) > 1 else []
        return GitCommitMessage(context=context, original=commit_msg_str, full=full, title=title, body=body)

    def __str__(self):
        return self.full

    def __eq__(self, other):
        return (
            isinstance(other, GitCommitMessage)
            and self.original == other.original
            and self.full == other.full
            and self.title == other.title
            and self.body == other.body
        )


class GitChangedFileStats:
    """Class representing the stats for a changed file in git"""

    def __init__(self, filepath, additions, deletions):
        self.filepath = Path(filepath)
        self.additions = additions
        self.deletions = deletions

    def __eq__(self, other):
        return (
            isinstance(other, GitChangedFileStats)
            and self.filepath == other.filepath
            and self.additions == other.additions
            and self.deletions == other.deletions
        )

    def __str__(self) -> str:
        return f"{self.filepath}: {self.additions} additions, {self.deletions} deletions"

    def __repr__(self) -> str:
        return (
            f'GitChangedFileStats(filepath="{self.filepath}", additions={self.additions}, deletions={self.deletions})'
        )


class GitCommit:
    """Class representing a git commit.
    A commit consists of: context, message, author name, author email, date, list of parent commit shas,
    list of changed files, list of branch names.
    In the context of gitlint, only the git context and commit message are required.
    """

    def __init__(
        self,
        context,
        message,
        sha=None,
        date=None,
        author_name=None,
        author_email=None,
        parents=None,
        changed_files_stats=None,
        branches=None,
    ):
        self.context = context
        self.message = message
        self.sha = sha
        self.date = date
        self.author_name = author_name
        self.author_email = author_email
        self.parents = parents or []  # parent commit hashes
        self.changed_files_stats = changed_files_stats or {}
        self.branches = branches or []

    @property
    def is_merge_commit(self):
        return self.message.title.startswith("Merge")

    @property
    def is_fixup_commit(self):
        return self.message.title.startswith("fixup!")

    @property
    def is_squash_commit(self):
        return self.message.title.startswith("squash!")

    @property
    def is_fixup_amend_commit(self):
        return self.message.title.startswith("amend!")

    @property
    def is_revert_commit(self):
        return self.message.title.startswith("Revert")

    @property
    def changed_files(self):
        return list(self.changed_files_stats.keys())

    def __str__(self):
        date_str = arrow.get(self.date).format(GIT_TIMEFORMAT) if self.date else None

        if len(self.changed_files_stats) > 0:
            changed_files_stats_str = "\n  " + "\n  ".join([str(stats) for stats in self.changed_files_stats.values()])
        else:
            changed_files_stats_str = " {}"

        return (
            f"--- Commit Message ----\n{self.message}\n"
            "--- Meta info ---------\n"
            f"Author: {self.author_name} <{self.author_email}>\n"
            f"Date:   {date_str}\n"
            f"is-merge-commit:  {self.is_merge_commit}\n"
            f"is-fixup-commit:  {self.is_fixup_commit}\n"
            f"is-fixup-amend-commit: {self.is_fixup_amend_commit}\n"
            f"is-squash-commit: {self.is_squash_commit}\n"
            f"is-revert-commit: {self.is_revert_commit}\n"
            f"Parents: {self.parents}\n"
            f"Branches: {self.branches}\n"
            f"Changed Files: {self.changed_files}\n"
            f"Changed Files Stats:{changed_files_stats_str}\n"
            "-----------------------"
        )

    def __eq__(self, other):
        # skip checking the context as context refers back to this obj, this will trigger a cyclic dependency
        return (
            isinstance(other, GitCommit)
            and self.message == other.message
            and self.sha == other.sha
            and self.author_name == other.author_name
            and self.author_email == other.author_email
            and self.date == other.date
            and self.parents == other.parents
            and self.is_merge_commit == other.is_merge_commit
            and self.is_fixup_commit == other.is_fixup_commit
            and self.is_fixup_amend_commit == other.is_fixup_amend_commit
            and self.is_squash_commit == other.is_squash_commit
            and self.is_revert_commit == other.is_revert_commit
            and self.changed_files == other.changed_files
            and self.changed_files_stats == other.changed_files_stats
            and self.branches == other.branches
        )


class LocalGitCommit(GitCommit, PropertyCache):
    """Class representing a git commit that exists in the local git repository.
    This class uses lazy loading: it defers reading information from the local git repository until the associated
    property is accessed for the first time. Properties are then cached for subsequent access.

    This approach ensures that we don't do 'expensive' git calls when certain properties are not actually used.
    In addition, reading the required info when it's needed rather than up front avoids adding delay during gitlint
    startup time and reduces gitlint's memory footprint.
    """

    def __init__(self, context, sha):
        PropertyCache.__init__(self)
        self.context = context
        self.sha = sha

    def _log(self):
        """Does a call to `git log` to determine a bunch of information about the commit."""
        long_format = "--pretty=%aN%x00%aE%x00%ai%x00%P%n%B"
        raw_commit = _git("log", self.sha, "-1", long_format, _cwd=self.context.repository_path).split("\n")

        (name, email, date, parents), commit_msg = raw_commit[0].split("\x00"), "\n".join(raw_commit[1:])

        commit_parents = [] if parents == "" else parents.split(" ")
        commit_is_merge_commit = len(commit_parents) > 1

        # "YYYY-MM-DD HH:mm:ss Z" -> ISO 8601-like format
        # Use arrow for datetime parsing, because apparently python is quirky around ISO-8601 dates:
        # http://stackoverflow.com/a/30696682/381010
        commit_date = arrow.get(date, GIT_TIMEFORMAT).datetime

        # Create Git commit object with the retrieved info
        commit_msg_obj = GitCommitMessage.from_full_message(self.context, commit_msg)

        self._cache.update(
            {
                "message": commit_msg_obj,
                "author_name": name,
                "author_email": email,
                "date": commit_date,
                "parents": commit_parents,
                "is_merge_commit": commit_is_merge_commit,
            }
        )

    @property
    def message(self):
        return self._try_cache("message", self._log)

    @property
    def author_name(self):
        return self._try_cache("author_name", self._log)

    @property
    def author_email(self):
        return self._try_cache("author_email", self._log)

    @property
    def date(self):
        return self._try_cache("date", self._log)

    @property
    def parents(self):
        return self._try_cache("parents", self._log)

    @property
    def branches(self):
        def cache_branches():
            # We have to parse 'git branch --contains <sha>' instead of 'git for-each-ref' to be compatible with
            # git versions < 2.7.0
            # https://stackoverflow.com/questions/45173979/can-i-force-git-branch-contains-tag-to-not-print-the-asterisk
            branches = _git("branch", "--contains", self.sha, _cwd=self.context.repository_path).split("\n")

            # This means that we need to remove any leading * that indicates the current branch. Note that we can
            # safely do this since git branches cannot contain '*' anywhere, so if we find an '*' we know it's output
            # from the git CLI and not part of the branch name. See https://git-scm.com/docs/git-check-ref-format
            # We also drop the last empty line from the output.
            self._cache["branches"] = [branch.replace("*", "").strip() for branch in branches[:-1]]

        return self._try_cache("branches", cache_branches)

    @property
    def is_merge_commit(self):
        return self._try_cache("is_merge_commit", self._log)

    @property
    def changed_files_stats(self):
        def cache_changed_files_stats():
            changed_files_stats_raw = _git(
                "diff-tree", "--no-commit-id", "--numstat", "-r", "--root", self.sha, _cwd=self.context.repository_path
            )
            self._cache["changed_files_stats"] = _parse_git_changed_file_stats(changed_files_stats_raw)

        return self._try_cache("changed_files_stats", cache_changed_files_stats)


class StagedLocalGitCommit(GitCommit, PropertyCache):
    """Class representing a git commit that has been staged, but not committed.

    Other than the commit message itself (and changed files), a lot of information is actually not known at staging
    time, since the commit hasn't happened yet. However, we can make educated guesses based on existing repository
    information.
    """

    def __init__(self, context, commit_message):
        PropertyCache.__init__(self)
        self.context = context
        self.message = commit_message
        self.sha = None
        self.parents = []  # Not really possible to determine before a commit

    @property
    @cache
    def author_name(self):
        try:
            return _git("config", "--get", "user.name", _cwd=self.context.repository_path).strip()
        except GitExitCodeError as e:
            raise GitContextError("Missing git configuration: please set user.name") from e

    @property
    @cache
    def author_email(self):
        try:
            return _git("config", "--get", "user.email", _cwd=self.context.repository_path).strip()
        except GitExitCodeError as e:
            raise GitContextError("Missing git configuration: please set user.email") from e

    @property
    @cache
    def date(self):
        # We don't know the actual commit date yet, but we make a pragmatic trade-off here by providing the current date
        # We get current date from arrow, reformat in git date format, then re-interpret it as a date.
        # This ensure we capture the same precision and timezone information that git does.
        return arrow.get(arrow.now().format(GIT_TIMEFORMAT), GIT_TIMEFORMAT).datetime

    @property
    @cache
    def branches(self):
        # We don't know the branch this commit will be part of yet, but we're pragmatic here and just return the
        # current branch, as for all intents and purposes, this will be what the user is looking for.
        return [self.context.current_branch]

    @property
    def changed_files_stats(self):
        def cache_changed_files_stats():
            changed_files_stats_raw = _git("diff", "--staged", "--numstat", "-r", _cwd=self.context.repository_path)
            self._cache["changed_files_stats"] = _parse_git_changed_file_stats(changed_files_stats_raw)

        return self._try_cache("changed_files_stats", cache_changed_files_stats)


class GitContext(PropertyCache):
    """Class representing the git context in which gitlint is operating: a data object storing information about
    the git repository that gitlint is linting.
    """

    def __init__(self, repository_path=None):
        PropertyCache.__init__(self)
        self.commits = []
        self.repository_path = repository_path

    @property
    @cache
    def commentchar(self):
        return git_commentchar(self.repository_path)

    @property
    @cache
    def current_branch(self):
        try:
            current_branch = _git("rev-parse", "--abbrev-ref", "HEAD", _cwd=self.repository_path).strip()
        except GitContextError:
            # Maybe there is no commit.  Try another way to get current branch (need Git 2.22+)
            current_branch = _git("branch", "--show-current", _cwd=self.repository_path).strip()
        return current_branch

    @staticmethod
    def from_commit_msg(commit_msg_str):
        """Determines git context based on a commit message.
        :param commit_msg_str: Full git commit message.
        """
        context = GitContext()
        commit_msg_obj = GitCommitMessage.from_full_message(context, commit_msg_str)
        commit = GitCommit(context, commit_msg_obj)
        context.commits.append(commit)
        return context

    @staticmethod
    def from_staged_commit(commit_msg_str, repository_path):
        """Determines git context based on a commit message that is a staged commit for a local git repository.
        :param commit_msg_str: Full git commit message.
        :param repository_path: Path to the git repository to retrieve the context from
        """
        context = GitContext(repository_path=repository_path)
        commit_msg_obj = GitCommitMessage.from_full_message(context, commit_msg_str)
        commit = StagedLocalGitCommit(context, commit_msg_obj)
        context.commits.append(commit)
        return context

    @staticmethod
    def from_local_repository(repository_path, refspec=None, commit_hashes=None):
        """Retrieves the git context from a local git repository.
        :param repository_path: Path to the git repository to retrieve the context from
        :param refspec: The commit(s) to retrieve (mutually exclusive with `commit_hash`)
        :param commit_hash: Hash of the commit to retrieve (mutually exclusive with `refspec`)
        """

        context = GitContext(repository_path=repository_path)

        if refspec:
            sha_list = _git("rev-list", refspec, _cwd=repository_path).split()
        elif commit_hashes:  # One or more commit hashes, just pass it to `git log -1`
            # Even though we have already been passed the commit hash, we ask git to retrieve this hash and
            # return it to us. This way we verify that the passed hash is a valid hash for the target repo and we
            # also convert it to the full hash format (we might have been passed a short hash).
            sha_list = []
            for commit_hash in commit_hashes:
                sha_list.append(_git("log", "-1", commit_hash, "--pretty=%H", _cwd=repository_path).replace("\n", ""))
        else:  # If no refspec is defined, fallback to the last commit on the current branch
            # We tried many things here e.g.: defaulting to e.g. HEAD or HEAD^... (incl. dealing with
            # repos that only have a single commit - HEAD^... doesn't work there), but then we still get into
            # problems with e.g. merge commits. Easiest solution is just taking the SHA from `git log -1`.
            sha_list = [_git("log", "-1", "--pretty=%H", _cwd=repository_path).replace("\n", "")]

        for sha in sha_list:
            commit = LocalGitCommit(context, sha)
            context.commits.append(commit)

        return context

    def __eq__(self, other):
        return (
            isinstance(other, GitContext)
            and self.commits == other.commits
            and self.repository_path == other.repository_path
            and self.commentchar == other.commentchar
            and self.current_branch == other.current_branch
        )



import logging

LOG = logging.getLogger("gitlint.deprecated")
DEPRECATED_LOG_FORMAT = "%(levelname)s: %(message)s"


class Deprecation:
    """Singleton class that handles deprecation warnings and behavior."""

    # LintConfig class that is used to determine deprecation behavior
    config = None

    # Set of warning messages that have already been logged, to prevent duplicate warnings
    warning_msgs = set()

    @classmethod
    def get_regex_method(cls, rule, regex_option):
        """Returns the regex method to be used for a given rule based on general.regex-style-search option.
        Logs a warning if the deprecated re.match method is returned."""

        # if general.regex-style-search is set, just return re.search
        if cls.config.regex_style_search:
            return regex_option.value.search

        warning_msg = (
            f"{rule.id} - {rule.name}: gitlint will be switching from using Python regex 'match' (match beginning) to "
            "'search' (match anywhere) semantics. "
            f"Please review your {rule.name}.regex option accordingly. "
            "To remove this warning, set general.regex-style-search=True. "
            "More details: https://jorisroovers.github.io/gitlint/configuration/#regex-style-search"
        )

        # Only log warnings once
        if warning_msg not in cls.warning_msgs:
            log = logging.getLogger("gitlint.deprecated.regex_style_search")
            log.warning(warning_msg)
        cls.warning_msgs.add(warning_msg)

        return regex_option.value.match



import logging

from gitlint import display
from gitlint import rules as gitlint_rules
from gitlint.deprecation import Deprecation

LOG = logging.getLogger(__name__)
logging.basicConfig()


class GitLinter:
    """Main linter class. This is where rules actually get applied. See the lint() method."""

    def __init__(self, config):
        self.config = config

        self.display = display.Display(config)

    def should_ignore_rule(self, rule):
        """Determines whether a rule should be ignored based on the general list of commits to ignore"""
        return rule.id in self.config.ignore or rule.name in self.config.ignore

    @property
    def configuration_rules(self):
        return [
            rule
            for rule in self.config.rules
            if isinstance(rule, gitlint_rules.ConfigurationRule) and not self.should_ignore_rule(rule)
        ]

    @property
    def title_line_rules(self):
        return [
            rule
            for rule in self.config.rules
            if isinstance(rule, gitlint_rules.LineRule)
            and rule.target == gitlint_rules.CommitMessageTitle
            and not self.should_ignore_rule(rule)
        ]

    @property
    def body_line_rules(self):
        return [
            rule
            for rule in self.config.rules
            if isinstance(rule, gitlint_rules.LineRule)
            and rule.target == gitlint_rules.CommitMessageBody
            and not self.should_ignore_rule(rule)
        ]

    @property
    def commit_rules(self):
        return [
            rule
            for rule in self.config.rules
            if isinstance(rule, gitlint_rules.CommitRule) and not self.should_ignore_rule(rule)
        ]

    @staticmethod
    def _apply_line_rules(lines, commit, rules, line_nr_start):
        """Iterates over the lines in a given list of lines and validates a given list of rules against each line"""
        all_violations = []
        line_nr = line_nr_start
        for line in lines:
            for rule in rules:
                violations = rule.validate(line, commit)
                if violations:
                    for violation in violations:
                        violation.line_nr = line_nr
                        all_violations.append(violation)
            line_nr += 1
        return all_violations

    @staticmethod
    def _apply_commit_rules(rules, commit):
        """Applies a set of rules against a given commit and gitcontext"""
        all_violations = []
        for rule in rules:
            violations = rule.validate(commit)
            if violations:
                all_violations.extend(violations)
        return all_violations

    def lint(self, commit):
        """Lint the last commit in a given git context by applying all ignore, title, body and commit rules."""
        LOG.debug("Linting commit %s", commit.sha or "[SHA UNKNOWN]")
        LOG.debug("Commit Object\n" + str(commit))

        # Ensure the Deprecation class has a reference to the config currently being used
        Deprecation.config = self.config

        # Apply config rules
        for rule in self.configuration_rules:
            rule.apply(self.config, commit)

        # Skip linting if this is a special commit type that is configured to be ignored
        ignore_commit_types = ["merge", "squash", "fixup", "fixup_amend", "revert"]
        for commit_type in ignore_commit_types:
            if getattr(commit, f"is_{commit_type}_commit") and getattr(self.config, f"ignore_{commit_type}_commits"):
                return []

        violations = []
        # determine violations by applying all rules
        violations.extend(self._apply_line_rules([commit.message.title], commit, self.title_line_rules, 1))
        violations.extend(self._apply_line_rules(commit.message.body, commit, self.body_line_rules, 2))
        violations.extend(self._apply_commit_rules(self.commit_rules, commit))

        # Sort violations by line number and rule_id. If there's no line nr specified (=common certain commit rules),
        # we replace None with -1 so that it always get's placed first. Note that we need this to do this to support
        # python 3, as None is not allowed in a list that is being sorted.
        violations.sort(key=lambda v: (-1 if v.line_nr is None else v.line_nr, v.rule_id))
        return violations

    def print_violations(self, violations):
        """Print a given set of violations to the standard error output"""
        for v in violations:
            line_nr = v.line_nr if v.line_nr else "-"
            self.display.e(f"{line_nr}: {v.rule_id}", exact=True)
            self.display.ee(f"{line_nr}: {v.rule_id} {v.message}", exact=True)
            if v.content:
                self.display.eee(f'{line_nr}: {v.rule_id} {v.message}: "{v.content}"', exact=True)
            else:
                self.display.eee(f"{line_nr}: {v.rule_id} {v.message}", exact=True)



import fnmatch
import importlib
import inspect
import os
import sys

from gitlint import options, rules


def find_rule_classes(extra_path):
    """
    Searches a given directory or python module for rule classes. This is done by
    adding the directory path to the python path, importing the modules and then finding
    any Rule class in those modules.

    :param extra_path: absolute directory or file path to search for rule classes
    :return: The list of rule classes that are found in the given directory or module
    """

    files = []
    modules = []

    if os.path.isfile(extra_path):
        files = [os.path.basename(extra_path)]
        directory = os.path.dirname(extra_path)
    elif os.path.isdir(extra_path):
        files = os.listdir(extra_path)
        directory = extra_path
    else:
        raise rules.UserRuleError(f"Invalid extra-path: {extra_path}")

    # Filter out files that are not python modules
    for filename in files:
        if fnmatch.fnmatch(filename, "*.py"):
            # We have to treat __init__ files a bit special: add the parent dir instead of the filename, and also
            # add their parent dir to the sys.path (this fixes import issues with pypy2).
            if filename == "__init__.py":
                modules.append(os.path.basename(directory))
                sys.path.append(os.path.dirname(directory))
            else:
                modules.append(os.path.splitext(filename)[0])

    # No need to continue if there are no modules specified
    if not modules:
        return []

    # Append the extra rules path to python path so that we can import them
    sys.path.append(directory)

    # Find all the rule classes in the found python files
    rule_classes = []
    for module in modules:
        # Import the module
        try:
            importlib.import_module(module)

        except Exception as e:
            raise rules.UserRuleError(f"Error while importing extra-path module '{module}': {e}") from e

        # Find all rule classes in the module. We do this my inspecting all members of the module and checking
        # 1) is it a class, if not, skip
        # 2) is the parent path the current module. If not, we are dealing with an imported class, skip
        # 3) is it a subclass of rule
        rule_classes.extend(
            [
                clazz
                for _, clazz in inspect.getmembers(sys.modules[module])
                if inspect.isclass(clazz)  # check isclass to ensure clazz.__module__ exists
                and clazz.__module__ == module  # ignore imported classes
                and (issubclass(clazz, (rules.LineRule, rules.CommitRule, rules.ConfigurationRule)))
            ]
        )

        # validate that the rule classes are valid user-defined rules
        for rule_class in rule_classes:
            assert_valid_rule_class(rule_class)

    return rule_classes


def assert_valid_rule_class(clazz, rule_type="User-defined"):  # noqa: PLR0912 (too many branches)
    """
    Asserts that a given rule clazz is valid by checking a number of its properties:
     - Rules must extend from  LineRule, CommitRule or ConfigurationRule
     - Rule classes must have id and name string attributes.
       The options_spec is optional, but if set, it must be a list of gitlint Options.
     - Rule classes must have a validate method. In case of a CommitRule, validate must take a single commit parameter.
       In case of LineRule, validate must take line and commit as first and second parameters.
     - LineRule classes must have a target class attributes that is set to either
    - ConfigurationRule classes must have an apply method that take `config` and `commit` as parameters.
       CommitMessageTitle or CommitMessageBody.
     - Rule id's cannot start with R, T, B, M or I as these rule ids are reserved for gitlint itself.
    """

    # Rules must extend from LineRule, CommitRule or ConfigurationRule
    if not issubclass(clazz, (rules.LineRule, rules.CommitRule, rules.ConfigurationRule)):
        msg = (
            f"{rule_type} rule class '{clazz.__name__}' "
            f"must extend from {rules.CommitRule.__module__}.{rules.LineRule.__name__}, "
            f"{rules.CommitRule.__module__}.{rules.CommitRule.__name__} or "
            f"{rules.CommitRule.__module__}.{rules.ConfigurationRule.__name__}"
        )
        raise rules.UserRuleError(msg)

    # Rules must have an id attribute
    if not hasattr(clazz, "id") or clazz.id is None or not clazz.id:
        raise rules.UserRuleError(f"{rule_type} rule class '{clazz.__name__}' must have an 'id' attribute")

    # Rule id's cannot start with gitlint reserved letters
    if clazz.id[0].upper() in ["R", "T", "B", "M", "I"]:
        msg = f"The id '{clazz.id[0]}' of '{clazz.__name__}' is invalid. Gitlint reserves ids starting with R,T,B,M,I"
        raise rules.UserRuleError(msg)

    # Rules must have a name attribute
    if not hasattr(clazz, "name") or clazz.name is None or not clazz.name:
        raise rules.UserRuleError(f"{rule_type} rule class '{clazz.__name__}' must have a 'name' attribute")

    # if set, options_spec must be a list of RuleOption
    if not isinstance(clazz.options_spec, list):
        msg = (
            f"The options_spec attribute of {rule_type.lower()} rule class '{clazz.__name__}' "
            f"must be a list of {options.RuleOption.__module__}.{options.RuleOption.__name__}"
        )
        raise rules.UserRuleError(msg)

    # check that all items in options_spec are actual gitlint options
    for option in clazz.options_spec:
        if not isinstance(option, options.RuleOption):
            msg = (
                f"The options_spec attribute of {rule_type.lower()} rule class '{clazz.__name__}' "
                f"must be a list of {options.RuleOption.__module__}.{options.RuleOption.__name__}"
            )
            raise rules.UserRuleError(msg)

    # Line/Commit rules must have a `validate` method
    # We use isroutine() as it's both python 2 and 3 compatible. Details: http://stackoverflow.com/a/17019998/381010
    if issubclass(clazz, (rules.LineRule, rules.CommitRule)):
        if not hasattr(clazz, "validate") or not inspect.isroutine(clazz.validate):
            raise rules.UserRuleError(f"{rule_type} rule class '{clazz.__name__}' must have a 'validate' method")
    # Configuration rules must have an `apply` method
    elif issubclass(clazz, rules.ConfigurationRule):  # noqa: SIM102
        if not hasattr(clazz, "apply") or not inspect.isroutine(clazz.apply):
            msg = f"{rule_type} Configuration rule class '{clazz.__name__}' must have an 'apply' method"
            raise rules.UserRuleError(msg)

    # LineRules must have a valid target: rules.CommitMessageTitle or rules.CommitMessageBody
    if issubclass(clazz, rules.LineRule):  # noqa: SIM102
        if clazz.target not in [rules.CommitMessageTitle, rules.CommitMessageBody]:
            msg = (
                f"The target attribute of the {rule_type.lower()} LineRule class '{clazz.__name__}' "
                f"must be either {rules.CommitMessageTitle.__module__}.{rules.CommitMessageTitle.__name__} "
                f"or {rules.CommitMessageTitle.__module__}.{rules.CommitMessageBody.__name__}"
            )
            raise rules.UserRuleError(msg)



import codecs
import locale
import os
import platform

# Note: While we can easily inline the logic related to the constants set in this module, we deliberately create
# small functions that encapsulate that logic as this enables easy unit testing. In particular, by creating functions
# we can easily mock the dependencies during testing, which is not possible if the code is not enclosed in a function
# and just executed at import-time.

########################################################################################################################
LOG_FORMAT = "%(levelname)s: %(name)s %(message)s"

########################################################################################################################
# PLATFORM_IS_WINDOWS


def platform_is_windows():
    return "windows" in platform.system().lower()


PLATFORM_IS_WINDOWS = platform_is_windows()

########################################################################################################################
# USE_SH_LIB
# Determine whether to use the `sh` library
# On windows we won't want to use the sh library since it's not supported - instead we'll use our own shell module.
# However, we want to be able to overwrite this behavior for testing using the GITLINT_USE_SH_LIB env var.


def use_sh_library():
    gitlint_use_sh_lib_env = os.environ.get("GITLINT_USE_SH_LIB", None)
    if gitlint_use_sh_lib_env:
        return gitlint_use_sh_lib_env == "1"
    return False


USE_SH_LIB = use_sh_library()

########################################################################################################################
# TERMINAL_ENCODING
# Encoding used for terminal encoding/decoding.


def getpreferredencoding():
    """Modified version of local.getpreferredencoding() that takes into account LC_ALL, LC_CTYPE, LANG env vars
    on windows and falls back to UTF-8."""
    fallback_encoding = "UTF-8"
    preferred_encoding = locale.getpreferredencoding() or fallback_encoding

    # On Windows, we mimic git/linux by trying to read the LC_ALL, LC_CTYPE, LANG env vars manually
    # (on Linux/MacOS the `getpreferredencoding()` call will take care of this).
    # We fallback to UTF-8
    if PLATFORM_IS_WINDOWS:
        preferred_encoding = fallback_encoding
        for env_var in ["LC_ALL", "LC_CTYPE", "LANG"]:
            encoding = os.environ.get(env_var, False)
            if encoding:
                # Support dotted (C.UTF-8) and non-dotted (C or UTF-8) charsets:
                # If encoding contains a dot: split and use second part, otherwise use everything
                dot_index = encoding.find(".")
                preferred_encoding = encoding[dot_index + 1 :] if dot_index != -1 else encoding
                break

        # We've determined what encoding the user *wants*, let's now check if it's actually a valid encoding on the
        # system. If not, fallback to UTF-8.
        # This scenario is fairly common on Windows where git sets LC_CTYPE=C when invoking the commit-msg hook, which
        # is not a valid encoding in Python on Windows.
        try:
            codecs.lookup(preferred_encoding)
        except LookupError:
            preferred_encoding = fallback_encoding

    return preferred_encoding


TERMINAL_ENCODING = getpreferredencoding()

########################################################################################################################
# FILE_ENCODING
# Gitlint assumes UTF-8 encoding for all file operations:
#  - reading/writing its own hook and config files
#  - reading/writing git commit messages
# Git does have i18n.commitEncoding and i18n.logOutputEncoding options which we might want to take into account,
# but that's not supported today.

FILE_ENCODING = "UTF-8"



"""
This module implements a shim for the 'sh' library, mainly for use on Windows (sh is not supported on Windows).
We might consider removing the 'sh' dependency altogether in the future, but 'sh' does provide a few
capabilities wrt dealing with more edge-case environments on *nix systems that are useful.
"""

import subprocess

from gitlint.utils import TERMINAL_ENCODING, USE_SH_LIB


def shell(cmd):
    """Convenience function that opens a given command in a shell. Does not use 'sh' library."""
    with subprocess.Popen(cmd, shell=True) as p:
        p.communicate()


if USE_SH_LIB:
    # import exceptions separately, this makes it a little easier to mock them out in the unit tests
    from sh import (
        CommandNotFound,
        ErrorReturnCode,
        git,
    )
else:

    class CommandNotFound(Exception):
        """Exception indicating a command was not found during execution"""

    class ShResult:
        """Result wrapper class. We use this to more easily migrate from using https://amoffat.github.io/sh/ to using
        the builtin subprocess module"""

        def __init__(self, full_cmd, stdout, stderr="", exitcode=0):
            self.full_cmd = full_cmd
            self.stdout = stdout
            self.stderr = stderr
            self.exit_code = exitcode

        def __str__(self):
            return self.stdout

    class ErrorReturnCode(ShResult, Exception):
        """ShResult subclass for unexpected results (acts as an exception)."""

    def git(*command_parts, **kwargs):
        """Git shell wrapper.
        Implemented as separate function here, so we can do a 'sh' style imports:
        `from shell import git`
        """
        args = ["git", *list(command_parts)]
        return _exec(*args, **kwargs)

    def _exec(*args, **kwargs):
        pipe = subprocess.PIPE
        popen_kwargs = {"stdout": pipe, "stderr": pipe, "shell": kwargs.get("_tty_out", False)}
        if "_cwd" in kwargs:
            popen_kwargs["cwd"] = kwargs["_cwd"]

        try:
            with subprocess.Popen(args, **popen_kwargs) as p:
                result = p.communicate()
        except FileNotFoundError as e:
            raise CommandNotFound from e

        exit_code = p.returncode
        stdout = result[0].decode(TERMINAL_ENCODING)
        stderr = result[1]  # 'sh' does not decode the stderr bytes to unicode
        full_cmd = "" if args is None else " ".join(args)

        # If not _ok_code is specified, then only a 0 exit code is allowed
        ok_exit_codes = kwargs.get("_ok_code", [0])

        if exit_code in ok_exit_codes:
            return ShResult(full_cmd, stdout, stderr, exit_code)

        # Unexpected error code => raise ErrorReturnCode
        raise ErrorReturnCode(full_cmd, stdout, stderr, p.returncode)



import os
import shutil
import stat

from gitlint.exception import GitlintError
from gitlint.git import git_hooks_dir
from gitlint.utils import FILE_ENCODING

COMMIT_MSG_HOOK_SRC_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "files", "commit-msg")
COMMIT_MSG_HOOK_DST_PATH = "commit-msg"
GITLINT_HOOK_IDENTIFIER = "### gitlint commit-msg hook start ###\n"


class GitHookInstallerError(GitlintError):
    pass


class GitHookInstaller:
    """Utility class that provides methods for installing and uninstalling the gitlint commitmsg hook."""

    @staticmethod
    def commit_msg_hook_path(lint_config):
        return os.path.join(git_hooks_dir(lint_config.target), COMMIT_MSG_HOOK_DST_PATH)

    @staticmethod
    def _assert_git_repo(target):
        """Asserts that a given target directory is a git repository"""
        hooks_dir = git_hooks_dir(target)
        if not os.path.isdir(hooks_dir):
            raise GitHookInstallerError(f"{target} is not a git repository.")

    @staticmethod
    def install_commit_msg_hook(lint_config):
        GitHookInstaller._assert_git_repo(lint_config.target)
        dest_path = GitHookInstaller.commit_msg_hook_path(lint_config)
        if os.path.exists(dest_path):
            raise GitHookInstallerError(
                f"There is already a commit-msg hook file present in {dest_path}.\n"
                "gitlint currently does not support appending to an existing commit-msg file."
            )

        # copy hook file
        shutil.copy(COMMIT_MSG_HOOK_SRC_PATH, dest_path)
        # make hook executable
        st = os.stat(dest_path)
        os.chmod(dest_path, st.st_mode | stat.S_IEXEC)

    @staticmethod
    def uninstall_commit_msg_hook(lint_config):
        GitHookInstaller._assert_git_repo(lint_config.target)
        dest_path = GitHookInstaller.commit_msg_hook_path(lint_config)
        if not os.path.exists(dest_path):
            raise GitHookInstallerError(f"There is no commit-msg hook present in {dest_path}.")

        with open(dest_path, encoding=FILE_ENCODING) as fp:
            lines = fp.readlines()
            if len(lines) < 2 or lines[1] != GITLINT_HOOK_IDENTIFIER:  # noqa: PLR2004 (Magic value used in comparison)
                msg = (
                    f"The commit-msg hook in {dest_path} was not installed by gitlint (or it was modified).\n"
                    "Uninstallation of 3th party or modified gitlint hooks is not supported."
                )
                raise GitHookInstallerError(msg)

        # If we are sure it's a gitlint hook, go ahead and remove it
        os.remove(dest_path)



import copy
import os
import re
import shutil
from collections import OrderedDict
from configparser import ConfigParser
from configparser import Error as ConfigParserError

from gitlint import (
    options,
    rule_finder,
    rules,
)
from gitlint.contrib import rules as contrib_rules
from gitlint.exception import GitlintError
from gitlint.utils import FILE_ENCODING


def handle_option_error(func):
    """Decorator that calls given method/function and handles any RuleOptionError gracefully by converting it to a
    LintConfigError."""

    def wrapped(*args):
        try:
            return func(*args)
        except options.RuleOptionError as e:
            raise LintConfigError(str(e)) from e

    return wrapped


class LintConfigError(GitlintError):
    pass


class LintConfig:
    """Class representing gitlint configuration.
    Contains active config as well as number of methods to easily get/set the config.
    """

    # Default tuple of rule classes (tuple because immutable).
    default_rule_classes = (
        rules.IgnoreByTitle,
        rules.IgnoreByBody,
        rules.IgnoreBodyLines,
        rules.IgnoreByAuthorName,
        rules.TitleMaxLength,
        rules.TitleTrailingWhitespace,
        rules.TitleLeadingWhitespace,
        rules.TitleTrailingPunctuation,
        rules.TitleHardTab,
        rules.TitleMustNotContainWord,
        rules.TitleRegexMatches,
        rules.TitleMinLength,
        rules.BodyMaxLineLength,
        rules.BodyMinLength,
        rules.BodyMissing,
        rules.BodyTrailingWhitespace,
        rules.BodyHardTab,
        rules.BodyFirstLineEmpty,
        rules.BodyChangedFileMention,
        rules.BodyRegexMatches,
        rules.AuthorValidEmail,
    )

    def __init__(self):
        self.rules = RuleCollection(self.default_rule_classes)
        self._verbosity = options.IntOption("verbosity", 3, "Verbosity")
        self._ignore_merge_commits = options.BoolOption("ignore-merge-commits", True, "Ignore merge commits")
        self._ignore_fixup_commits = options.BoolOption("ignore-fixup-commits", True, "Ignore fixup commits")
        self._ignore_fixup_amend_commits = options.BoolOption(
            "ignore-fixup-amend-commits", True, "Ignore fixup amend commits"
        )
        self._ignore_squash_commits = options.BoolOption("ignore-squash-commits", True, "Ignore squash commits")
        self._ignore_revert_commits = options.BoolOption("ignore-revert-commits", True, "Ignore revert commits")
        self._debug = options.BoolOption("debug", False, "Enable debug mode")
        self._extra_path = None
        target_description = "Path of the target git repository (default=current working directory)"
        self._target = options.PathOption("target", os.path.realpath(os.getcwd()), target_description)
        self._ignore = options.ListOption("ignore", [], "List of rule-ids to ignore")
        self._contrib = options.ListOption("contrib", [], "List of contrib-rules to enable")
        self._config_path = None
        ignore_stdin_description = "Ignore any stdin data. Useful for running in CI server."
        self._ignore_stdin = options.BoolOption("ignore-stdin", False, ignore_stdin_description)
        self._staged = options.BoolOption("staged", False, "Read staged commit meta-info from the local repository.")
        self._fail_without_commits = options.BoolOption(
            "fail-without-commits", False, "Hard fail when the target commit range is empty"
        )
        self._regex_style_search = options.BoolOption(
            "regex-style-search", False, "Use `search` instead of `match` semantics for regex rules"
        )

    @property
    def target(self):
        return self._target.value if self._target else None

    @target.setter
    @handle_option_error
    def target(self, value):
        return self._target.set(value)

    @property
    def verbosity(self):
        return self._verbosity.value

    @verbosity.setter
    @handle_option_error
    def verbosity(self, value):
        self._verbosity.set(value)
        if self.verbosity < 0 or self.verbosity > 3:  # noqa: PLR2004 (Magic value used in comparison)
            raise LintConfigError("Option 'verbosity' must be set between 0 and 3")

    @property
    def ignore_merge_commits(self):
        return self._ignore_merge_commits.value

    @ignore_merge_commits.setter
    @handle_option_error
    def ignore_merge_commits(self, value):
        return self._ignore_merge_commits.set(value)

    @property
    def ignore_fixup_commits(self):
        return self._ignore_fixup_commits.value

    @ignore_fixup_commits.setter
    @handle_option_error
    def ignore_fixup_commits(self, value):
        return self._ignore_fixup_commits.set(value)

    @property
    def ignore_fixup_amend_commits(self):
        return self._ignore_fixup_amend_commits.value

    @ignore_fixup_amend_commits.setter
    @handle_option_error
    def ignore_fixup_amend_commits(self, value):
        return self._ignore_fixup_amend_commits.set(value)

    @property
    def ignore_squash_commits(self):
        return self._ignore_squash_commits.value

    @ignore_squash_commits.setter
    @handle_option_error
    def ignore_squash_commits(self, value):
        return self._ignore_squash_commits.set(value)

    @property
    def ignore_revert_commits(self):
        return self._ignore_revert_commits.value

    @ignore_revert_commits.setter
    @handle_option_error
    def ignore_revert_commits(self, value):
        return self._ignore_revert_commits.set(value)

    @property
    def debug(self):
        return self._debug.value

    @debug.setter
    @handle_option_error
    def debug(self, value):
        return self._debug.set(value)

    @property
    def ignore(self):
        return self._ignore.value

    @ignore.setter
    def ignore(self, value):
        if value == "all":
            value = [rule.id for rule in self.rules]
        return self._ignore.set(value)

    @property
    def ignore_stdin(self):
        return self._ignore_stdin.value

    @ignore_stdin.setter
    @handle_option_error
    def ignore_stdin(self, value):
        return self._ignore_stdin.set(value)

    @property
    def staged(self):
        return self._staged.value

    @staged.setter
    @handle_option_error
    def staged(self, value):
        return self._staged.set(value)

    @property
    def fail_without_commits(self):
        return self._fail_without_commits.value

    @fail_without_commits.setter
    @handle_option_error
    def fail_without_commits(self, value):
        return self._fail_without_commits.set(value)

    @property
    def regex_style_search(self):
        return self._regex_style_search.value

    @regex_style_search.setter
    @handle_option_error
    def regex_style_search(self, value):
        return self._regex_style_search.set(value)

    @property
    def extra_path(self):
        return self._extra_path.value if self._extra_path else None

    @extra_path.setter
    def extra_path(self, value):
        try:
            if self.extra_path:
                self._extra_path.set(value)
            else:
                self._extra_path = options.PathOption(
                    "extra-path", value, "Path to a directory or module with extra user-defined rules", type="both"
                )

            # Make sure we unload any previously loaded extra-path rules
            self.rules.delete_rules_by_attr("is_user_defined", True)

            # Find rules in the new extra-path and add them to the existing rules
            rule_classes = rule_finder.find_rule_classes(self.extra_path)
            self.rules.add_rules(rule_classes, {"is_user_defined": True})

        except (options.RuleOptionError, rules.UserRuleError) as e:
            raise LintConfigError(str(e)) from e

    @property
    def contrib(self):
        return self._contrib.value

    @contrib.setter
    def contrib(self, value):
        try:
            self._contrib.set(value)

            # Make sure we unload any previously loaded contrib rules when re-setting the value
            self.rules.delete_rules_by_attr("is_contrib", True)

            # Load all classes from the contrib directory
            contrib_dir_path = os.path.dirname(os.path.realpath(contrib_rules.__file__))
            rule_classes = rule_finder.find_rule_classes(contrib_dir_path)

            # For each specified contrib rule, check whether it exists among the contrib classes
            for rule_id_or_name in self.contrib:
                rule_class = next((rc for rc in rule_classes if rule_id_or_name in (rc.id, rc.name)), False)

                # If contrib rule exists, instantiate it and add it to the rules list
                if rule_class:
                    self.rules.add_rule(rule_class, rule_class.id, {"is_contrib": True})
                else:
                    raise LintConfigError(f"No contrib rule with id or name '{rule_id_or_name}' found.")

        except (options.RuleOptionError, rules.UserRuleError) as e:
            raise LintConfigError(str(e)) from e

    def _get_option(self, rule_name_or_id, option_name):
        rule = self.rules.find_rule(rule_name_or_id)
        if not rule:
            raise LintConfigError(f"No such rule '{rule_name_or_id}'")

        option = rule.options.get(option_name)
        if not option:
            raise LintConfigError(f"Rule '{rule_name_or_id}' has no option '{option_name}'")

        return option

    def get_rule_option(self, rule_name_or_id, option_name):
        """Returns the value of a given option for a given rule. LintConfigErrors will be raised if the
        rule or option don't exist."""
        option = self._get_option(rule_name_or_id, option_name)
        return option.value

    def set_rule_option(self, rule_name_or_id, option_name, option_value):
        """Attempts to set a given value for a given option for a given rule.
        LintConfigErrors will be raised if the rule or option don't exist or if the value is invalid."""
        option = self._get_option(rule_name_or_id, option_name)
        try:
            option.set(option_value)
        except options.RuleOptionError as e:
            msg = f"'{option_value}' is not a valid value for option '{rule_name_or_id}.{option_name}'. {e}."
            raise LintConfigError(msg) from e

    def set_general_option(self, option_name, option_value):
        attr_name = option_name.replace("-", "_")
        # only allow setting general options that exist and don't start with an underscore
        if not hasattr(self, attr_name) or attr_name[0] == "_":
            raise LintConfigError(f"'{option_name}' is not a valid gitlint option")

        # else
        setattr(self, attr_name, option_value)

    def __eq__(self, other):
        return (
            isinstance(other, LintConfig)
            and self.rules == other.rules
            and self.verbosity == other.verbosity
            and self.target == other.target
            and self.extra_path == other.extra_path
            and self.contrib == other.contrib
            and self.ignore_merge_commits == other.ignore_merge_commits
            and self.ignore_fixup_commits == other.ignore_fixup_commits
            and self.ignore_fixup_amend_commits == other.ignore_fixup_amend_commits
            and self.ignore_squash_commits == other.ignore_squash_commits
            and self.ignore_revert_commits == other.ignore_revert_commits
            and self.ignore_stdin == other.ignore_stdin
            and self.staged == other.staged
            and self.fail_without_commits == other.fail_without_commits
            and self.regex_style_search == other.regex_style_search
            and self.debug == other.debug
            and self.ignore == other.ignore
            and self._config_path == other._config_path
        )

    def __str__(self):
        # config-path is not a user exposed variable, so don't print it under the general section
        return (
            f"config-path: {self._config_path}\n"
            "[GENERAL]\n"
            f"extra-path: {self.extra_path}\n"
            f"contrib: {self.contrib}\n"
            f"ignore: {','.join(self.ignore)}\n"
            f"ignore-merge-commits: {self.ignore_merge_commits}\n"
            f"ignore-fixup-commits: {self.ignore_fixup_commits}\n"
            f"ignore-fixup-amend-commits: {self.ignore_fixup_amend_commits}\n"
            f"ignore-squash-commits: {self.ignore_squash_commits}\n"
            f"ignore-revert-commits: {self.ignore_revert_commits}\n"
            f"ignore-stdin: {self.ignore_stdin}\n"
            f"staged: {self.staged}\n"
            f"fail-without-commits: {self.fail_without_commits}\n"
            f"regex-style-search: {self.regex_style_search}\n"
            f"verbosity: {self.verbosity}\n"
            f"debug: {self.debug}\n"
            f"target: {self.target}\n"
            f"[RULES]\n{self.rules}"
        )


class RuleCollection:
    """Class representing an ordered list of rules. Methods are provided to easily retrieve, add or delete rules."""

    def __init__(self, rule_classes=None, rule_attrs=None):
        # Use an ordered dict so that the order in which rules are applied is always the same
        self._rules = OrderedDict()
        if rule_classes:
            self.add_rules(rule_classes, rule_attrs)

    def find_rule(self, rule_id_or_name):
        rule = self._rules.get(rule_id_or_name)
        # if not found, try finding rule by name
        if not rule:
            rule = next((rule for rule in self._rules.values() if rule.name == rule_id_or_name), None)
        return rule

    def add_rule(self, rule_class, rule_id, rule_attrs=None):
        """Instantiates and adds a rule to RuleCollection.
        Note: There can be multiple instantiations of the same rule_class in the RuleCollection, as long as the
        rule_id is unique.
        :param rule_class python class representing the rule
        :param rule_id unique identifier for the rule. If not unique, it will
                       overwrite the existing rule with that id
        :param rule_attrs dictionary of attributes to set on the instantiated rule obj
        """
        rule_obj = rule_class()
        rule_obj.id = rule_id
        if rule_attrs:
            for key, val in rule_attrs.items():
                setattr(rule_obj, key, val)
        self._rules[rule_obj.id] = rule_obj

    def add_rules(self, rule_classes, rule_attrs=None):
        """Convenience method to add multiple rules at once based on a list of rule classes."""
        for rule_class in rule_classes:
            self.add_rule(rule_class, rule_class.id, rule_attrs)

    def delete_rules_by_attr(self, attr_name, attr_val):
        """Deletes all rules from the collection that match a given attribute name and value"""
        # Create a new list based on _rules.values() because in python 3, values() is a ValuesView as opposed to a list
        # This means you can't modify the ValueView while iterating over it.
        for rule in list(self._rules.values()):
            if hasattr(rule, attr_name) and (getattr(rule, attr_name) == attr_val):
                del self._rules[rule.id]

    def __iter__(self):
        yield from self._rules.values()

    def __eq__(self, other):
        return isinstance(other, RuleCollection) and self._rules == other._rules

    def __len__(self):
        return len(self._rules)

    def __str__(self):
        return_str = ""
        for rule in self._rules.values():
            return_str += f"  {rule.id}: {rule.name}\n"
            for option_name, option_value in sorted(rule.options.items()):
                if option_value.value is None:
                    option_val_repr = None
                elif isinstance(option_value.value, list):
                    option_val_repr = ",".join(option_value.value)
                elif isinstance(option_value, options.RegexOption):
                    option_val_repr = option_value.value.pattern
                else:
                    option_val_repr = option_value.value
                return_str += f"     {option_name}={option_val_repr}\n"
        return return_str


class LintConfigBuilder:
    """Factory class that can build gitlint config.
    This is primarily useful to deal with complex configuration scenarios where configuration can be set and overridden
    from various sources (typically according to certain precedence rules) before the actual config should be
    normalized, validated and build. Example usage can be found in gitlint.cli.
    """

    RULE_QUALIFIER_SYMBOL = ":"

    def __init__(self):
        self._config_blueprint = OrderedDict()
        self._config_path = None

    def set_option(self, section, option_name, option_value):
        if section not in self._config_blueprint:
            self._config_blueprint[section] = OrderedDict()
        self._config_blueprint[section][option_name] = option_value

    def set_config_from_commit(self, commit):
        """Given a git commit, applies config specified in the commit message.
        Supported:
         - gitlint-ignore: all
        """
        for line in commit.message.body:
            pattern = re.compile(r"^gitlint-ignore:\s*(.*)")
            matches = pattern.match(line)
            if matches and len(matches.groups()) == 1:
                self.set_option("general", "ignore", matches.group(1))

    def set_config_from_string_list(self, config_options):
        """Given a list of config options of the form "<rule>.<option>=<value>", parses out the correct rule and option
        and sets the value accordingly in this factory object."""
        for config_option in config_options:
            try:
                config_name, option_value = config_option.split("=", 1)
                if not option_value:
                    raise ValueError()
                rule_name, option_name = config_name.split(".", 1)
                self.set_option(rule_name, option_name, option_value)
            except ValueError as e:  # raised if the config string is invalid
                raise LintConfigError(
                    f"'{config_option}' is an invalid configuration option. Use '<rule>.<option>=<value>'"
                ) from e

    def set_from_config_file(self, filename):
        """Loads lint config from an ini-style config file"""
        if not os.path.exists(filename):
            raise LintConfigError(f"Invalid file path: {filename}")
        self._config_path = os.path.realpath(filename)
        try:
            parser = ConfigParser()

            with open(filename, encoding=FILE_ENCODING) as config_file:
                parser.read_file(config_file, filename)

            for section_name in parser.sections():
                for option_name, option_value in parser.items(section_name):
                    self.set_option(section_name, option_name, str(option_value))

        except ConfigParserError as e:
            raise LintConfigError(str(e)) from e

    def _add_named_rule(self, config, qualified_rule_name):
        """Adds a Named Rule to a given LintConfig object.
        IMPORTANT: This method does *NOT* overwrite existing Named Rules with the same canonical id.
        """

        # Split up named rule in its parts: the name/id that specifies the parent rule,
        # And the name of the rule instance itself
        rule_name_parts = qualified_rule_name.split(self.RULE_QUALIFIER_SYMBOL, 1)
        rule_name = rule_name_parts[1].strip()
        parent_rule_specifier = rule_name_parts[0].strip()

        # assert that the rule name is valid:
        # - not empty
        # - no whitespace or colons
        if rule_name == "" or bool(re.search("\\s|:", rule_name, re.UNICODE)):
            msg = f"The rule-name part in '{qualified_rule_name}' cannot contain whitespace, colons or be empty"
            raise LintConfigError(msg)

        # find parent rule
        parent_rule = config.rules.find_rule(parent_rule_specifier)
        if not parent_rule:
            msg = f"No such rule '{parent_rule_specifier}' (named rule: '{qualified_rule_name}')"
            raise LintConfigError(msg)

        # Determine canonical id and name by recombining the parent id/name and instance name parts.
        canonical_id = parent_rule.__class__.id + self.RULE_QUALIFIER_SYMBOL + rule_name
        canonical_name = parent_rule.__class__.name + self.RULE_QUALIFIER_SYMBOL + rule_name

        # Add the rule to the collection of rules if it's not there already
        if not config.rules.find_rule(canonical_id):
            config.rules.add_rule(parent_rule.__class__, canonical_id, {"is_named": True, "name": canonical_name})

        return canonical_id

    def build(self, config=None):
        """Build a real LintConfig object by normalizing and validating the options that were previously set on this
        factory."""
        # If we are passed a config object, then rebuild that object instead of building a new lintconfig object from
        # scratch
        if not config:
            config = LintConfig()

        config._config_path = self._config_path

        # Set general options first as this might change the behavior or validity of the other options
        general_section = self._config_blueprint.get("general")
        if general_section:
            for option_name, option_value in general_section.items():
                config.set_general_option(option_name, option_value)

        for section_name, section_dict in self._config_blueprint.items():
            for option_name, option_value in section_dict.items():
                # Skip over the general section, as we've already done that above
                if section_name != "general":
                    # If the section name contains a colon (:), then this section is defining a Named Rule
                    # Which means we need to instantiate that Named Rule in the config.
                    if self.RULE_QUALIFIER_SYMBOL in section_name:
                        section_name = self._add_named_rule(config, section_name)

                    config.set_rule_option(section_name, option_name, option_value)

        return config

    def clone(self):
        """Creates an exact copy of a LintConfigBuilder."""
        builder = LintConfigBuilder()
        builder._config_blueprint = copy.deepcopy(self._config_blueprint)
        builder._config_path = self._config_path
        return builder


GITLINT_CONFIG_TEMPLATE_SRC_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "files/gitlint")


class LintConfigGenerator:
    @staticmethod
    def generate_config(dest):
        """Generates a gitlint config file at the given destination location.
        Expects that the given ```dest``` points to a valid destination."""
        shutil.copyfile(GITLINT_CONFIG_TEMPLATE_SRC_PATH, dest)



import re
from pathlib import Path
from typing import Tuple

from gitlint.rules import CommitRule, RuleViolation


class AllowedAuthors(CommitRule):
    """Enforce that only authors listed in the AUTHORS file are allowed to commit."""

    authors_file_names = ("AUTHORS", "AUTHORS.txt", "AUTHORS.md")
    parse_authors = re.compile(r"^(?P<name>.*) <(?P<email>.*)>$", re.MULTILINE)

    name = "contrib-allowed-authors"

    id = "CC3"

    @classmethod
    def _read_authors_from_file(cls, git_ctx) -> Tuple[str, str]:
        for file_name in cls.authors_file_names:
            path = Path(git_ctx.repository_path) / file_name
            if path.exists():
                authors_file = path
                break
        else:
            raise FileNotFoundError("No AUTHORS file found!")

        authors_file_content = authors_file.read_text("utf-8")
        authors = re.findall(cls.parse_authors, authors_file_content)

        return set(authors), authors_file.name

    def validate(self, commit):
        registered_authors, authors_file_name = AllowedAuthors._read_authors_from_file(commit.message.context)

        author = (commit.author_name, commit.author_email.lower())

        if author not in registered_authors:
            return [
                RuleViolation(
                    self.id,
                    f"Author not in '{authors_file_name}' file: " f'"{commit.author_name} <{commit.author_email}>"',
                )
            ]
        return []
